@@ -1,6 +1,8 @@
+import enum
 import os
 import sys
 import traceback
+from collections import defaultdict
 from enum import Enum
 from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union
 
@@ -49,14 +51,27 @@ TARGETS_TYPE = click.Choice((*ALL_TARGETS_NAMES, "all"))
 
 
 def register_target(function: Target) -> Target:
-    """Register a new testing target for schemathesis CLI."""
+    """Register a new testing target for schemathesis CLI.
+
+    :param function: A function that will be called to calculate a metric passed to ``hypothesis.target``.
+    """
     targets_module.ALL_TARGETS += (function,)
     TARGETS_TYPE.choices += (function.__name__,)  # type: ignore
     return function
 
 
 def register_check(function: CheckFunction) -> CheckFunction:
-    """Register a new check for schemathesis CLI."""
+    """Register a new check for schemathesis CLI.
+
+    :param function: A function to validate API responses.
+
+    .. code-block:: python
+
+        @schemathesis.register_check
+        def new_check(response, case):
+            # some awesome assertions!
+            ...
+    """
     checks_module.ALL_CHECKS += (function,)
     CHECKS_TYPE.choices += (function.__name__,)  # type: ignore
     return function
@@ -101,10 +116,55 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
         load_hook(pre_run)
 
 
-@schemathesis.command(short_help="Perform schemathesis test.")
+class ParameterGroup(enum.Enum):
+    filtering = "Filtering", "These options define what parts of the API will be tested."
+    validation = "Validation", "Options, responsible for how responses & schemas will be checked."
+    hypothesis = "Hypothesis", "Configuration of the underlying Hypothesis engine."
+    generic = "Generic", None
+
+
+class CommandWithCustomHelp(click.Command):
+    def format_options(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
+        # Group options first
+        groups = defaultdict(list)
+        for param in self.get_params(ctx):
+            rv = param.get_help_record(ctx)
+            if rv is not None:
+                if isinstance(param, GroupedOption):
+                    group = param.group
+                else:
+                    group = ParameterGroup.generic
+                groups[group].append(rv)
+        # Then display groups separately with optional description
+        for group in ParameterGroup:
+            opts = groups[group]
+            group_name, description = group.value
+            with formatter.section(f"{group_name} options"):
+                if description:
+                    formatter.write_paragraph()
+                    formatter.write_text(description)
+                    formatter.write_paragraph()
+                formatter.write_dl(opts)
+
+
+class GroupedOption(click.Option):
+    def __init__(self, *args: Any, group: ParameterGroup, **kwargs: Any):
+        super().__init__(*args, **kwargs)
+        self.group = group
+
+
+@schemathesis.command(short_help="Perform schemathesis test.", cls=CommandWithCustomHelp)
 @click.argument("schema", type=str, callback=callbacks.validate_schema)
 @click.option(
-    "--checks", "-c", multiple=True, help="List of checks to run.", type=CHECKS_TYPE, default=DEFAULT_CHECKS_NAMES
+    "--checks",
+    "-c",
+    multiple=True,
+    help="List of checks to run.",
+    type=CHECKS_TYPE,
+    default=DEFAULT_CHECKS_NAMES,
+    cls=GroupedOption,
+    group=ParameterGroup.validation,
+    show_default=True,
 )
 @click.option(
     "--data-generation-method",
@@ -113,11 +173,14 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     help="Defines how Schemathesis generates data for tests.",
     type=CSVOption(DataGenerationMethod),
     default=DataGenerationMethod.default(),
+    show_default=True,
 )
 @click.option(
     "--max-response-time",
     help="A custom check that will fail if the response time is greater than the specified one in milliseconds.",
     type=click.IntRange(min=1),
+    cls=GroupedOption,
+    group=ParameterGroup.validation,
 )
 @click.option(
     "--target",
@@ -127,9 +190,24 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     help="Targets for input generation.",
     type=TARGETS_TYPE,
     default=DEFAULT_TARGETS_NAMES,
+    show_default=True,
 )
 @click.option(
-    "-x", "--exitfirst", "exit_first", is_flag=True, default=False, help="Exit instantly on first error or failed test."
+    "-x",
+    "--exitfirst",
+    "exit_first",
+    is_flag=True,
+    default=False,
+    help="Exit instantly on first error or failed test.",
+    show_default=True,
+)
+@click.option(
+    "--dry-run",
+    "dry_run",
+    is_flag=True,
+    default=False,
+    help="Disable sending data to the application and checking responses. "
+    "Helpful to verify whether data is generated at all.",
 )
 @click.option(
     "--auth", "-a", help="Server user and password. Example: USER:PASSWORD", type=str, callback=callbacks.validate_auth
@@ -140,6 +218,7 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     type=click.Choice(["basic", "digest"], case_sensitive=False),
     default="basic",
     help="The authentication mechanism to be used. Defaults to 'basic'.",
+    show_default=True,
 )
 @click.option(
     "--header",
@@ -156,8 +235,10 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     "endpoints",
     type=str,
     multiple=True,
-    help=r"Filter schemathesis test by endpoint pattern. Example: users/\d+",
+    help=r"Filter schemathesis tests by API operation path pattern. Example: users/\d+",
     callback=callbacks.validate_regex,
+    cls=GroupedOption,
+    group=ParameterGroup.filtering,
 )
 @click.option(
     "--method",
@@ -165,8 +246,10 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     "methods",
     type=str,
     multiple=True,
-    help="Filter schemathesis test by HTTP method.",
+    help="Filter schemathesis tests by HTTP method.",
     callback=callbacks.validate_regex,
+    cls=GroupedOption,
+    group=ParameterGroup.filtering,
 )
 @click.option(
     "--tag",
@@ -174,8 +257,10 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     "tags",
     type=str,
     multiple=True,
-    help="Filter schemathesis test by schema tag pattern.",
+    help="Filter schemathesis tests by schema tag pattern.",
     callback=callbacks.validate_regex,
+    cls=GroupedOption,
+    group=ParameterGroup.filtering,
 )
 @click.option(
     "--operation-id",
@@ -183,8 +268,10 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     "operation_ids",
     type=str,
     multiple=True,
-    help="Filter schemathesis test by operationId pattern.",
+    help="Filter schemathesis tests by operationId pattern.",
     callback=callbacks.validate_regex,
+    cls=GroupedOption,
+    group=ParameterGroup.filtering,
 )
 @click.option(
     "--workers",
@@ -196,6 +283,7 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
         choices_repr=f"[auto|{MIN_WORKERS}-{MAX_WORKERS}]",
     ),
     default=str(DEFAULT_WORKERS),
+    show_default=True,
     callback=callbacks.convert_workers,
 )
 @click.option(
@@ -217,15 +305,27 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     "You can also pass the path to a CA_BUNDLE file for private certs.",
     type=str,
     default="true",
+    show_default=True,
     callback=callbacks.convert_request_tls_verify,
 )
-@click.option("--validate-schema", help="Enable or disable validation of input schema.", type=bool, default=True)
 @click.option(
-    "--skip-deprecated-endpoints",
-    help="Skip testing of deprecated endpoints.",
+    "--validate-schema",
+    help="Enable or disable validation of input schema.",
+    type=bool,
+    default=True,
+    show_default=True,
+    cls=GroupedOption,
+    group=ParameterGroup.validation,
+)
+@click.option(
+    "--skip-deprecated-operations",
+    help="Skip testing of deprecated API operations.",
     is_flag=True,
     is_eager=True,
     default=False,
+    show_default=True,
+    cls=GroupedOption,
+    group=ParameterGroup.filtering,
 )
 @click.option(
     "--junit-xml", help="Create junit-xml style report file at given path.", type=click.File("w", encoding="utf-8")
@@ -236,6 +336,7 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     is_flag=True,
     is_eager=True,
     default=False,
+    show_default=True,
 )
 @click.option(
     "--store-network-log", help="Store requests and responses into a file.", type=click.File("w", encoding="utf-8")
@@ -256,9 +357,10 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     "--stateful-recursion-limit",
     help="Limit recursion depth for stateful testing.",
     default=DEFAULT_STATEFUL_RECURSION_LIMIT,
+    show_default=True,
     type=click.IntRange(1, 100),
     cls=DeprecatedOption,
-    removed_in="3.0",
+    removed_in="4.0",
 )
 @click.option(
     "--force-schema-version",
@@ -270,28 +372,60 @@ def schemathesis(pre_run: Optional[str] = None) -> None:
     help="Duration in milliseconds that each individual example with a test is not allowed to exceed.",
     # max value to avoid overflow. It is the maximum amount of days in milliseconds
     type=OptionalInt(1, 999999999 * 24 * 3600 * 1000),
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
+)
+@click.option(
+    "--hypothesis-derandomize",
+    help="Use Hypothesis's deterministic mode.",
+    is_flag=True,
+    default=None,
+    show_default=True,
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
 )
-@click.option("--hypothesis-derandomize", help="Use Hypothesis's deterministic mode.", is_flag=True, default=None)
 @click.option(
     "--hypothesis-max-examples",
-    help="Maximum number of generated examples per each method/endpoint combination.",
+    help="Maximum number of generated examples per each method/path combination.",
     type=click.IntRange(1),
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
 )
-@click.option("--hypothesis-phases", help="Control which phases should be run.", type=CSVOption(hypothesis.Phase))
 @click.option(
-    "--hypothesis-report-multiple-bugs", help="Raise only the exception with the smallest minimal example.", type=bool
+    "--hypothesis-phases",
+    help="Control which phases should be run.",
+    type=CSVOption(hypothesis.Phase),
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
+)
+@click.option(
+    "--hypothesis-report-multiple-bugs",
+    help="Raise only the exception with the smallest minimal example.",
+    type=bool,
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
+)
+@click.option(
+    "--hypothesis-seed",
+    help="Set a seed to use for all Hypothesis tests.",
+    type=int,
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
 )
-@click.option("--hypothesis-seed", help="Set a seed to use for all Hypothesis tests.", type=int)
 @click.option(
     "--hypothesis-suppress-health-check",
     help="Comma-separated list of health checks to disable.",
     type=CSVOption(hypothesis.HealthCheck),
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
 )
 @click.option(
     "--hypothesis-verbosity",
     help="Verbosity level of Hypothesis messages.",
     type=click.Choice([item.name for item in hypothesis.Verbosity]),
     callback=callbacks.convert_verbosity,
+    cls=GroupedOption,
+    group=ParameterGroup.hypothesis,
 )
 @click.option("--verbosity", "-v", help="Reduce verbosity of error output.", count=True)
 def run(
@@ -304,6 +438,7 @@ def run(
     max_response_time: Optional[int] = None,
     targets: Iterable[str] = DEFAULT_TARGETS_NAMES,
     exit_first: bool = False,
+    dry_run: bool = False,
     endpoints: Optional[Filter] = None,
     methods: Optional[Filter] = None,
     tags: Optional[Filter] = None,
@@ -314,7 +449,7 @@ def run(
     request_timeout: Optional[int] = None,
     request_tls_verify: bool = True,
     validate_schema: bool = True,
-    skip_deprecated_endpoints: bool = False,
+    skip_deprecated_operations: bool = False,
     junit_xml: Optional[click.utils.LazyFile] = None,
     show_errors_tracebacks: bool = False,
     store_network_log: Optional[click.utils.LazyFile] = None,
@@ -337,6 +472,7 @@ def run(
     SCHEMA must be a valid URL or file path pointing to an Open API / Swagger specification.
     """
     # pylint: disable=too-many-locals
+    check_auth(auth, headers)
     selected_targets = tuple(target for target in targets_module.ALL_TARGETS if target.__name__ in targets)
 
     if "all" in checks:
@@ -359,6 +495,7 @@ def run(
         app=app,
         seed=hypothesis_seed,
         exit_first=exit_first,
+        dry_run=dry_run,
         store_interactions=store_network_log is not None,
         checks=selected_checks,
         data_generation_methods=data_generation_methods,
@@ -366,7 +503,7 @@ def run(
         targets=selected_targets,
         workers_num=workers_num,
         validate_schema=validate_schema,
-        skip_deprecated_endpoints=skip_deprecated_endpoints,
+        skip_deprecated_operations=skip_deprecated_operations,
         fixups=fixups,
         stateful=stateful,
         stateful_recursion_limit=stateful_recursion_limit,
@@ -379,7 +516,14 @@ def run(
         hypothesis_suppress_health_check=hypothesis_suppress_health_check,
         hypothesis_verbosity=hypothesis_verbosity,
     )
-    execute(prepared_runner, workers_num, show_errors_tracebacks, store_network_log, junit_xml, verbosity)
+    execute(
+        prepared_runner, workers_num, show_errors_tracebacks, validate_schema, store_network_log, junit_xml, verbosity
+    )
+
+
+def check_auth(auth: Optional[Tuple[str, str]], headers: Dict[str, str]) -> None:
+    if auth is not None and "authorization" in {header.lower() for header in headers}:
+        raise click.BadParameter("Passing `--auth` together with `--header` that sets `Authorization` is not allowed.")
 
 
 def get_output_handler(workers_num: int) -> EventHandler:
@@ -413,6 +557,7 @@ def execute(
     prepared_runner: Generator[events.ExecutionEvent, None, None],
     workers_num: int,
     show_errors_tracebacks: bool,
+    validate_schema: bool,
     store_network_log: Optional[click.utils.LazyFile],
     junit_xml: Optional[click.utils.LazyFile],
     verbosity: int,
@@ -428,6 +573,7 @@ def execute(
     execution_context = ExecutionContext(
         workers_num=workers_num,
         show_errors_tracebacks=show_errors_tracebacks,
+        validate_schema=validate_schema,
         cassette_file_name=store_network_log.name if store_network_log is not None else None,
         junit_xml_file=junit_xml.name if junit_xml is not None else None,
         verbosity=verbosity,
@@ -437,8 +583,12 @@ def execute(
         for event in prepared_runner:
             for handler in handlers:
                 handler.handle_event(execution_context, event)
-    except click.exceptions.Exit:
-        raise
+            if isinstance(event, events.Finished):
+                if event.has_failures or event.has_errors:
+                    exit_code = 1
+                else:
+                    exit_code = 0
+                sys.exit(exit_code)
     except Exception as exc:
         for handler in handlers:
             handler.shutdown()
