@@ -1,4 +1,5 @@
 import base64
+import json
 import re
 import sys
 import threading
@@ -12,8 +13,8 @@ from requests.cookies import RequestsCookieJar
 from requests.structures import CaseInsensitiveDict
 
 from .. import constants
-from ..models import Interaction
 from ..runner import events
+from ..runner.serialization import SerializedCheck, SerializedInteraction
 from .context import ExecutionContext
 from .handlers import EventHandler
 
@@ -45,7 +46,12 @@ class CassetteWriter(EventHandler):
             # Seed is always present at this point, the original Optional[int] type is there because `TestResult`
             # instance is created before `seed` is generated on the hypothesis side
             seed = cast(int, event.result.seed)
-            self.queue.put(Process(status=event.status.name.upper(), seed=seed, interactions=event.result.interactions))
+            self.queue.put(
+                Process(
+                    seed=seed,
+                    interactions=event.result.interactions,
+                )
+            )
         if isinstance(event, events.Finished):
             self.shutdown()
 
@@ -66,9 +72,8 @@ class Initialize:
 class Process:
     """A new chunk of data should be processed."""
 
-    status: str = attr.ib()  # pragma: no mutate
     seed: int = attr.ib()  # pragma: no mutate
-    interactions: List[Interaction] = attr.ib()  # pragma: no mutate
+    interactions: List[SerializedInteraction] = attr.ib()  # pragma: no mutate
 
 
 @attr.s(slots=True)  # pragma: no mutate
@@ -99,11 +104,20 @@ def worker(file_handle: click.utils.LazyFile, queue: Queue) -> None:
     stream = file_handle.open()
 
     def format_header_values(values: List[str]) -> str:
-        return "\n".join(f"      - '{v}'" for v in values)
+        return "\n".join(f"      - {json.dumps(v)}" for v in values)
 
     def format_headers(headers: Dict[str, List[str]]) -> str:
         return "\n".join(f"      {name}:\n{format_header_values(values)}" for name, values in headers.items())
 
+    def format_check_message(message: Optional[str]) -> str:
+        return "~" if message is None else f"{repr(message)}"
+
+    def format_checks(checks: List[SerializedCheck]) -> str:
+        return "\n".join(
+            f"    - name: '{check.name}'\n      status: '{check.value.name.upper()}'\n      message: {format_check_message(check.message)}"
+            for check in checks
+        )
+
     while True:
         item = queue.get()
         if isinstance(item, Initialize):
@@ -114,12 +128,15 @@ http_interactions:"""
             )
         elif isinstance(item, Process):
             for interaction in item.interactions:
+                status = interaction.status.name.upper()
                 stream.write(
                     f"""\n- id: '{current_id}'
-  status: '{item.status}'
+  status: '{status}'
   seed: '{item.seed}'
   elapsed: '{interaction.response.elapsed}'
   recorded_at: '{interaction.recorded_at}'
+  checks:
+{format_checks(interaction.checks)}
   request:
     uri: '{interaction.request.uri}'
     method: '{interaction.request.method}'
@@ -131,7 +148,7 @@ http_interactions:"""
   response:
     status:
       code: '{interaction.response.status_code}'
-      message: '{interaction.response.message}'
+      message: {json.dumps(interaction.response.message)}
     headers:
 {format_headers(interaction.response.headers)}
     body:
