@@ -1,16 +1,19 @@
 from hashlib import sha1
 from json import JSONDecodeError
-from typing import Dict, Type, Union
+from typing import Any, Callable, Dict, NoReturn, Optional, Type, Union
 
 import attr
 from jsonschema import ValidationError
 
+from .constants import SERIALIZERS_SUGGESTION_MESSAGE
 from .utils import GenericResponse
 
 
 class CheckFailed(AssertionError):
     """Custom error type to distinguish from arbitrary AssertionError that may happen in the dependent libraries."""
 
+    __module__ = "builtins"
+
 
 CACHE: Dict[Union[str, int], Type[CheckFailed]] = {}
 
@@ -35,8 +38,8 @@ def _get_hashed_exception(prefix: str, message: str) -> Type[CheckFailed]:
 
 
 def get_grouped_exception(prefix: str, *exceptions: AssertionError) -> Type[CheckFailed]:
-    # The prefix is needed to distinguish multiple endpoints with the same error messages
-    # that are coming from different endpoints
+    # The prefix is needed to distinguish multiple operations with the same error messages
+    # that are coming from different operations
     messages = [exception.args[0] for exception in exceptions]
     message = "".join(messages)
     return _get_hashed_exception("GroupedException", f"{prefix}{message}")
@@ -79,8 +82,60 @@ def get_headers_error(message: str) -> Type[CheckFailed]:
     return _get_hashed_exception("MissingHeadersError", message)
 
 
+@attr.s(slots=True)
 class InvalidSchema(Exception):
-    """Schema associated with an endpoint contains an error."""
+    """Schema associated with an API operation contains an error."""
+
+    __module__ = "builtins"
+    message: Optional[str] = attr.ib(default=None)
+    path: Optional[str] = attr.ib(default=None)
+    method: Optional[str] = attr.ib(default=None)
+    full_path: Optional[str] = attr.ib(default=None)
+
+    def as_failing_test_function(self) -> Callable:
+        """Create a test function that will fail.
+
+        This approach allows us to use default pytest reporting style for operation-level schema errors.
+        """
+
+        def actual_test(*args: Any, **kwargs: Any) -> NoReturn:
+            __tracebackhide__ = True  # pylint: disable=unused-variable
+            raise self
+
+        return actual_test
+
+
+class NonCheckError(Exception):
+    """An error happened in side the runner, but is not related to failed checks.
+
+    Used primarily to not let Hypothesis to consider the test as flaky or detect multiple failures as we handle it
+    on our side.
+    """
+
+    __module__ = "builtins"
+
+
+SERIALIZATION_NOT_POSSIBLE_MESSAGE = (
+    f"Schemathesis can't serialize data to any of the defined media types: {{}} \n{SERIALIZERS_SUGGESTION_MESSAGE}"
+)
+
+
+class SerializationNotPossible(Exception):
+    """Not possible to serialize to any of the media types defined for some API operation.
+
+    Usually, there is still `application/json` along with less common ones, but this error happens when there is no
+    media type that Schemathesis knows how to serialize data to.
+    """
+
+    __module__ = "builtins"
+
+    @classmethod
+    def from_media_types(cls, *media_types: str) -> "SerializationNotPossible":
+        return cls(SERIALIZATION_NOT_POSSIBLE_MESSAGE.format(", ".join(media_types)))
+
+
+class InvalidRegularExpression(Exception):
+    __module__ = "builtins"
 
 
 @attr.s  # pragma: no mutate
