@@ -5,32 +5,47 @@ from collections import Counter
 from contextlib import contextmanager
 from copy import deepcopy
 from enum import IntEnum
+from itertools import chain
 from logging import LogRecord
-from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, Iterator, List, Optional, Sequence, Tuple, Union, cast
-from urllib.parse import urljoin, urlsplit, urlunsplit
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    Generic,
+    Iterator,
+    List,
+    NoReturn,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+    Union,
+    cast,
+)
+from urllib.parse import quote, unquote, urljoin, urlsplit, urlunsplit
 
 import attr
 import curlify
 import requests
 import werkzeug
+from hypothesis import event, note, reject
 from hypothesis.strategies import SearchStrategy
 from starlette.testclient import TestClient as ASGIClient
 
-from .constants import USER_AGENT, DataGenerationMethod
-from .exceptions import CheckFailed, InvalidSchema, get_grouped_exception
-from .types import Body, Cookies, FormData, Headers, PathParameters, Query
-from .utils import GenericResponse, WSGIResponse, get_response_payload
+from . import serializers
+from .constants import SERIALIZERS_SUGGESTION_MESSAGE, USER_AGENT, DataGenerationMethod
+from .exceptions import CheckFailed, InvalidSchema, SerializationNotPossible, get_grouped_exception
+from .parameters import Parameter, ParameterSet, PayloadAlternatives
+from .serializers import Serializer, SerializerContext
+from .types import Body, Cookies, FormData, Headers, NotSet, PathParameters, Query
+from .utils import NOT_SET, GenericResponse, WSGIResponse, deprecated_property, get_response_payload
 
 if TYPE_CHECKING:
     from .hooks import HookDispatcher
     from .schemas import BaseSchema
-    from .stateful import Feedback, Stateful, StatefulTest
-
-
-MISSING_STATEFUL_ARGUMENT_MESSAGE = (
-    "To use `store_response` you need to enable stateful testing by adding "
-    "`stateful=Stateful.links` to your `parametrize` call."
-)
+    from .stateful import Stateful, StatefulTest
 
 
 @attr.s(slots=True)  # pragma: no mutate
@@ -41,27 +56,37 @@ class CaseSource:
     response: GenericResponse = attr.ib()  # pragma: no mutate
 
 
+def cant_serialize(media_type: str) -> NoReturn:  # type: ignore
+    """Reject the current example if we don't know how to send this data to the application."""
+    event_text = f"Can't serialize data to `{media_type}`."
+    note(f"{event_text} {SERIALIZERS_SUGGESTION_MESSAGE}")
+    event(event_text)
+    reject()  # type: ignore
+
+
 @attr.s(slots=True, repr=False)  # pragma: no mutate
 class Case:  # pylint: disable=too-many-public-methods
     """A single test case parameters."""
 
-    endpoint: "Endpoint" = attr.ib()  # pragma: no mutate
+    operation: "APIOperation" = attr.ib()  # pragma: no mutate
     path_parameters: Optional[PathParameters] = attr.ib(default=None)  # pragma: no mutate
     headers: Optional[Headers] = attr.ib(default=None)  # pragma: no mutate
     cookies: Optional[Cookies] = attr.ib(default=None)  # pragma: no mutate
     query: Optional[Query] = attr.ib(default=None)  # pragma: no mutate
-    body: Optional[Body] = attr.ib(default=None)  # pragma: no mutate
-    form_data: Optional[FormData] = attr.ib(default=None)  # pragma: no mutate
+    # By default, there is no body, but we can't use `None` as the default value because it clashes with `null`
+    # which is a valid payload.
+    body: Union[Body, NotSet] = attr.ib(default=NOT_SET)  # pragma: no mutate
 
-    feedback: "Feedback" = attr.ib(default=None)  # pragma: no mutate
     source: Optional[CaseSource] = attr.ib(default=None)  # pragma: no mutate
+    # The media type for cases with a payload. For example, "application/json"
+    media_type: Optional[str] = attr.ib(default=None)  # pragma: no mutate
 
     def __repr__(self) -> str:
         parts = [f"{self.__class__.__name__}("]
         first = True
-        for name in ("path_parameters", "headers", "cookies", "query", "body", "form_data"):
+        for name in ("path_parameters", "headers", "cookies", "query", "body"):
             value = getattr(self, name)
-            if value is not None:
+            if value not in (None, NOT_SET):
                 if first:
                     first = False
                 else:
@@ -69,25 +94,29 @@ class Case:  # pylint: disable=too-many-public-methods
                 parts.extend((name, "=", repr(value)))
         return "".join(parts) + ")"
 
+    @deprecated_property(removed_in="4.0", replacement="operation")
+    def endpoint(self) -> "APIOperation":
+        return self.operation
+
     @property
     def path(self) -> str:
-        return self.endpoint.path
+        return self.operation.path
 
     @property
     def full_path(self) -> str:
-        return self.endpoint.full_path
+        return self.operation.full_path
 
     @property
     def method(self) -> str:
-        return self.endpoint.method.upper()
+        return self.operation.method.upper()
 
     @property
     def base_url(self) -> Optional[str]:
-        return self.endpoint.base_url
+        return self.operation.base_url
 
     @property
     def app(self) -> Any:
-        return self.endpoint.app
+        return self.operation.app
 
     def set_source(self, response: GenericResponse, case: "Case") -> None:
         self.source = CaseSource(case=case, response=response)
@@ -97,10 +126,11 @@ class Case:  # pylint: disable=too-many-public-methods
         # pylint: disable=not-a-mapping
         try:
             return self.path.format(**self.path_parameters or {})
-        except KeyError as variable:
-            raise InvalidSchema(
-                f"Missing path parameter {variable}. It either not defined or has no `required: true` in its definition"
-            ) from variable
+        except KeyError as exc:
+            # This may happen when a path template has a placeholder for variable "X", but parameter "X" is not defined
+            # in the parameters list.
+            # When `exc` is formatted, it is the missing key name in quotes. E.g. 'id'
+            raise InvalidSchema(f"Path parameter {exc} is not defined") from exc
 
     def get_full_base_url(self) -> Optional[str]:
         """Create a full base url, adding "localhost" for WSGI apps."""
@@ -121,15 +151,16 @@ class Case:  # pylint: disable=too-many-public-methods
             "Cookies": self.cookies,
             "Query": self.query,
             "Body": self.body,
-            "Form data": self.form_data,
         }
         max_length = max(map(len, output))
         template = f"{{:<{max_length}}} : {{}}"
-        return [
-            template.format(key, value)
-            for key, value in output.items()
-            if (key == "Body" and value is not None) or value not in (None, {})
-        ]
+
+        def should_display(key: str, value: Any) -> bool:
+            if key == "Body":
+                return value is not NOT_SET
+            return value is not None
+
+        return [template.format(key, value) for key, value in output.items() if should_display(key, value)]
 
     def get_code_to_reproduce(
         self, headers: Optional[Dict[str, Any]] = None, request: Optional[requests.PreparedRequest] = None
@@ -151,10 +182,14 @@ class Case:  # pylint: disable=too-many-public-methods
             kwargs["headers"] = final_headers
         method = kwargs["method"].lower()
 
+        def should_display(key: str, value: Any) -> bool:
+            if key in ("method", "url"):
+                return False
+            # Parameters are either absent because they are not defined or are optional
+            return value not in (None, {})
+
         printed_kwargs = ", ".join(
-            f"{key}={repr(value)}"
-            for key, value in kwargs.items()
-            if key not in ("method", "url") and value is not None
+            f"{key}={repr(value)}" for key, value in kwargs.items() if should_display(key, value)
         )
         args_repr = f"'{kwargs['url']}'"
         if printed_kwargs:
@@ -191,23 +226,36 @@ class Case:  # pylint: disable=too-many-public-methods
             final_headers["User-Agent"] = USER_AGENT
         return final_headers
 
+    def _get_serializer(self) -> Optional[Serializer]:
+        """Get a serializer for the payload, if there is any."""
+        if self.media_type is not None:
+            cls = serializers.get(self.media_type)
+            if cls is None:
+                all_media_types = self.operation.get_request_payload_content_types()
+                if all(serializers.get(media_type) is None for media_type in all_media_types):
+                    raise SerializationNotPossible.from_media_types(*all_media_types)
+                cant_serialize(self.media_type)
+            return cls()
+        return None
+
     def as_requests_kwargs(
         self, base_url: Optional[str] = None, headers: Optional[Dict[str, str]] = None
     ) -> Dict[str, Any]:
         """Convert the case into a dictionary acceptable by requests."""
         final_headers = self._get_headers(headers)
+        if self.media_type and self.media_type != "multipart/form-data" and self.body is not NOT_SET:
+            # `requests` will handle multipart form headers with the proper `boundary` value.
+            final_headers["Content-Type"] = self.media_type
         base_url = self._get_base_url(base_url)
         formatted_path = self.formatted_path.lstrip("/")  # pragma: no mutate
-        url = urljoin(base_url + "/", formatted_path)
-        # Form data and body are mutually exclusive
-        extra: Dict[str, Optional[Body]]
-        if self.form_data:
-            files, data = self.endpoint.prepare_multipart(self.form_data)
-            extra = {"files": files, "data": data}
-        elif is_multipart(self.body):
-            extra = {"data": self.body}
+        url = unquote(urljoin(base_url + "/", quote(formatted_path)))
+        extra: Dict[str, Any]
+        serializer = self._get_serializer()
+        if serializer is not None and self.body is not NOT_SET:
+            context = SerializerContext(case=self)
+            extra = serializer.as_requests(context, self.body)
         else:
-            extra = {"json": self.body}
+            extra = {}
         return {
             "method": self.method,
             "url": url,
@@ -235,31 +283,24 @@ class Case:  # pylint: disable=too-many-public-methods
         response = session.request(**data)  # type: ignore
         if close_session:
             session.close()
-        if self.feedback:
-            self.store_response(response)
         return response
 
-    def store_response(self, response: GenericResponse) -> None:
-        if self.feedback is None:
-            raise RuntimeError(MISSING_STATEFUL_ARGUMENT_MESSAGE)
-        self.feedback.add_test_case(self, response)
-
     def as_werkzeug_kwargs(self, headers: Optional[Dict[str, str]] = None) -> Dict[str, Any]:
         """Convert the case into a dictionary acceptable by werkzeug.Client."""
         final_headers = self._get_headers(headers)
-        extra: Dict[str, Optional[Body]]
-        if self.form_data:
-            extra = {"data": self.form_data}
-            final_headers = final_headers or {}
-            if "multipart/form-data" in self.endpoint.get_request_payload_content_types():
-                final_headers.setdefault("Content-Type", "multipart/form-data")
-        elif is_multipart(self.body):
-            extra = {"data": self.body}
+        if self.media_type and self.body is not NOT_SET:
+            # If we need to send a payload, then the Content-Type header should be set
+            final_headers["Content-Type"] = self.media_type
+        extra: Dict[str, Any]
+        serializer = self._get_serializer()
+        if serializer is not None and self.body is not NOT_SET:
+            context = SerializerContext(case=self)
+            extra = serializer.as_werkzeug(context, self.body)
         else:
-            extra = {"json": self.body}
+            extra = {}
         return {
             "method": self.method,
-            "path": self.endpoint.schema.get_full_path(self.formatted_path),
+            "path": self.operation.schema.get_full_path(self.formatted_path),
             "headers": final_headers,
             "query_string": self.query,
             **extra,
@@ -278,8 +319,6 @@ class Case:  # pylint: disable=too-many-public-methods
             response = client.open(**data, **kwargs)
         requests_kwargs = self.as_requests_kwargs(base_url=self.get_full_base_url(), headers=headers)
         response.request = requests.Request(**requests_kwargs).prepare()
-        if self.feedback:
-            self.store_response(response)
         return response
 
     def call_asgi(
@@ -305,18 +344,26 @@ class Case:  # pylint: disable=too-many-public-methods
         checks: Tuple["CheckFunction", ...] = (),
         additional_checks: Tuple["CheckFunction", ...] = (),
     ) -> None:
+        """Validate application response.
+
+        By default, all available checks will be applied.
+
+        :param response: Application response.
+        :param checks: A tuple of check functions that accept ``response`` and ``case``.
+        :param additional_checks: A tuple of additional checks that will be executed after ones from the ``checks``
+            argument.
+        """
         from .checks import ALL_CHECKS  # pylint: disable=import-outside-toplevel
 
         checks = checks or ALL_CHECKS
-        checks += additional_checks
         errors = []
-        for check in checks:
+        for check in chain(checks, additional_checks):
             try:
                 check(response, self)
             except CheckFailed as exc:
                 errors.append(exc)
         if errors:
-            exception_cls = get_grouped_exception(self.endpoint.verbose_name, *errors)
+            exception_cls = get_grouped_exception(self.operation.verbose_name, *errors)
             formatted_errors = "\n\n".join(f"{idx}. {error.args[0]}" for idx, error in enumerate(errors, 1))
             code = self.get_code_to_reproduce(request=response.request)
             payload = get_response_payload(response)
@@ -337,46 +384,24 @@ class Case:  # pylint: disable=too-many-public-methods
         self.validate_response(response, checks)
 
     def get_full_url(self) -> str:
-        """Make a full URL to the current endpoint, including query parameters."""
+        """Make a full URL to the current API operation, including query parameters."""
         base_url = self.base_url or "http://localhost"
         kwargs = self.as_requests_kwargs(base_url)
         request = requests.Request(**kwargs)
         prepared = requests.Session().prepare_request(request)  # type: ignore
-        return prepared.url  # type: ignore
+        return prepared.url
 
     def partial_deepcopy(self) -> "Case":
         return self.__class__(
-            endpoint=self.endpoint.partial_deepcopy(),
+            operation=self.operation.partial_deepcopy(),
             path_parameters=deepcopy(self.path_parameters),
             headers=deepcopy(self.headers),
             cookies=deepcopy(self.cookies),
             query=deepcopy(self.query),
             body=deepcopy(self.body),
-            form_data=deepcopy(self.form_data),
         )
 
 
-def is_multipart(item: Optional[Body]) -> bool:
-    """A poor detection if the body should be a multipart request.
-
-    It traverses the structure and if it contains bytes in any value, then it is a multipart request, because
-    it may happen only if there was `format: binary`, which usually is in multipart payloads.
-    Probably a better way would be checking actual content types defined in `requestBody` and drive behavior based on
-    that fact.
-    """
-    if isinstance(item, bytes):
-        return True
-    if isinstance(item, dict):
-        for value in item.values():
-            if is_multipart(value):
-                return True
-    if isinstance(item, list):
-        for value in item:
-            if is_multipart(value):
-                return True
-    return False
-
-
 @contextmanager
 def cookie_handler(client: werkzeug.Client, cookies: Optional[Cookies]) -> Generator[None, None, None]:
     """Set cookies required for a call."""
@@ -390,15 +415,11 @@ def cookie_handler(client: werkzeug.Client, cookies: Optional[Cookies]) -> Gener
             client.delete_cookie("localhost", key)
 
 
-def empty_object() -> Dict[str, Any]:
-    return {"properties": {}, "additionalProperties": False, "type": "object", "required": []}
-
-
 @attr.s(slots=True)  # pragma: no mutate
-class EndpointDefinition:
-    """A wrapper to store not resolved endpoint definitions.
+class OperationDefinition:
+    """A wrapper to store not resolved API operation definitions.
 
-    To prevent recursion errors we need to store definitions without resolving references. But endpoint definitions
+    To prevent recursion errors we need to store definitions without resolving references. But operation definitions
     itself can be behind a reference (when there is a ``$ref`` in ``paths`` values), therefore we need to store this
     scope change to have a proper reference resolving later.
     """
@@ -406,28 +427,39 @@ class EndpointDefinition:
     raw: Dict[str, Any] = attr.ib()  # pragma: no mutate
     resolved: Dict[str, Any] = attr.ib()  # pragma: no mutate
     scope: str = attr.ib()  # pragma: no mutate
-    parameters: List[Dict[str, Any]] = attr.ib()  # pragma: no mutate
+    parameters: Sequence[Parameter] = attr.ib()  # pragma: no mutate
 
 
-@attr.s(slots=True)  # pragma: no mutate
-class Endpoint:
-    """A container that could be used for test cases generation."""
+P = TypeVar("P", bound=Parameter)
+
+
+@attr.s  # pragma: no mutate
+class APIOperation(Generic[P]):
+    """A single operation defined in an API.
+
+    You can get one via a ``schema`` instance.
+
+    .. code-block:: python
+
+        # Get the POST /items operation
+        operation = schema["/items"]["POST"]
+
+    """
 
     # `path` does not contain `basePath`
     # Example <scheme>://<host>/<basePath>/users - "/users" is path
     # https://swagger.io/docs/specification/2-0/api-host-and-base-path/
     path: str = attr.ib()  # pragma: no mutate
     method: str = attr.ib()  # pragma: no mutate
-    definition: EndpointDefinition = attr.ib(repr=False)  # pragma: no mutate
+    definition: OperationDefinition = attr.ib(repr=False)  # pragma: no mutate
     schema: "BaseSchema" = attr.ib()  # pragma: no mutate
     app: Any = attr.ib(default=None)  # pragma: no mutate
     base_url: Optional[str] = attr.ib(default=None)  # pragma: no mutate
-    path_parameters: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
-    headers: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
-    cookies: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
-    query: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
-    body: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
-    form_data: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
+    path_parameters: ParameterSet[P] = attr.ib(factory=ParameterSet)  # pragma: no mutate
+    headers: ParameterSet[P] = attr.ib(factory=ParameterSet)  # pragma: no mutate
+    cookies: ParameterSet[P] = attr.ib(factory=ParameterSet)  # pragma: no mutate
+    query: ParameterSet[P] = attr.ib(factory=ParameterSet)  # pragma: no mutate
+    body: PayloadAlternatives[P] = attr.ib(factory=PayloadAlternatives)  # pragma: no mutate
 
     @property
     def verbose_name(self) -> str:
@@ -441,33 +473,57 @@ class Endpoint:
     def links(self) -> Dict[str, Dict[str, Any]]:
         return self.schema.get_links(self)
 
+    def add_parameter(self, parameter: P) -> None:
+        """Add a new processed parameter to an API operation.
+
+        :param parameter: A parameter that will be used with this operation.
+        :rtype: None
+        """
+        lookup_table = {
+            "path": self.path_parameters,
+            "header": self.headers,
+            "cookie": self.cookies,
+            "query": self.query,
+            "body": self.body,
+        }
+        # If the parameter has a typo, then by default, there will be an error from `jsonschema` earlier.
+        # But if the user wants to skip schema validation, we choose to ignore a malformed parameter.
+        # In this case, we still might generate some tests for an API operation, but without this parameter,
+        # which is better than skip the whole operation from testing.
+        if parameter.location in lookup_table:
+            container = lookup_table[parameter.location]
+            container.add(parameter)
+
     def as_strategy(
         self,
         hooks: Optional["HookDispatcher"] = None,
-        feedback: Optional["Feedback"] = None,
         data_generation_method: DataGenerationMethod = DataGenerationMethod.default(),
     ) -> SearchStrategy:
-        return self.schema.get_case_strategy(self, hooks, feedback, data_generation_method)
+        """Turn this API operation into a Hypothesis strategy."""
+        return self.schema.get_case_strategy(self, hooks, data_generation_method)
 
     def get_strategies_from_examples(self) -> List[SearchStrategy[Case]]:
-        """Get examples from the endpoint."""
+        """Get examples from the API operation."""
         return self.schema.get_strategies_from_examples(self)
 
     def get_stateful_tests(self, response: GenericResponse, stateful: Optional["Stateful"]) -> Sequence["StatefulTest"]:
         return self.schema.get_stateful_tests(response, self, stateful)
 
-    def get_hypothesis_conversions(self, location: str) -> Optional[Callable]:
-        return self.schema.get_hypothesis_conversion(self, location)
+    def get_parameter_serializer(self, location: str) -> Optional[Callable]:
+        """Get a function that serializes parameters for the given location.
 
-    def prepare_multipart(self, form_data: Optional[FormData]) -> Tuple[Optional[List], Optional[Dict[str, Any]]]:
-        if not form_data:
-            return None, None
+        It handles serializing data into various `collectionFormat` options and similar.
+        Note that payload is handled by this function - it is handled by serializers.
+        """
+        return self.schema.get_parameter_serializer(self, location)
+
+    def prepare_multipart(self, form_data: FormData) -> Tuple[Optional[List], Optional[Dict[str, Any]]]:
         return self.schema.prepare_multipart(form_data, self)
 
     def get_request_payload_content_types(self) -> List[str]:
         return self.schema.get_request_payload_content_types(self)
 
-    def partial_deepcopy(self) -> "Endpoint":
+    def partial_deepcopy(self) -> "APIOperation":
         return self.__class__(
             path=self.path,  # string, immutable
             method=self.method,  # string, immutable
@@ -480,11 +536,10 @@ class Endpoint:
             cookies=deepcopy(self.cookies),
             query=deepcopy(self.query),
             body=deepcopy(self.body),
-            form_data=deepcopy(self.form_data),
         )
 
-    def clone(self, **components: Any) -> "Endpoint":
-        """Create a new instance of this endpoint with updated components."""
+    def clone(self, **components: Any) -> "APIOperation":
+        """Create a new instance of this API operation with updated components."""
         return self.__class__(
             path=self.path,
             method=self.method,
@@ -497,7 +552,6 @@ class Endpoint:
             headers=components["headers"],
             cookies=components["cookies"],
             body=components["body"],
-            form_data=self.form_data,
         )
 
     def make_case(
@@ -507,17 +561,16 @@ class Endpoint:
         headers: Optional[Headers] = None,
         cookies: Optional[Cookies] = None,
         query: Optional[Query] = None,
-        body: Optional[Body] = None,
-        form_data: Optional[FormData] = None,
+        body: Union[Body, NotSet] = NOT_SET,
     ) -> Case:
+        """Create a new example for this API operation."""
         return Case(
-            endpoint=self,
+            operation=self,
             path_parameters=path_parameters,
             headers=headers,
             cookies=cookies,
             query=query,
             body=body,
-            form_data=form_data,
         )
 
     @property
@@ -526,9 +579,14 @@ class Endpoint:
         return f"#/paths/{path}/{self.method}"
 
     def validate_response(self, response: GenericResponse) -> None:
+        """Validate API response for conformance.
+
+        :raises CheckFailed: If the response does not conform to the API schema.
+        """
         return self.schema.validate_response(self, response)
 
     def is_response_valid(self, response: GenericResponse) -> bool:
+        """Validate API response for conformance."""
         try:
             self.validate_response(response)
             return True
@@ -536,6 +594,10 @@ class Endpoint:
             return False
 
 
+# backward-compatibility
+Endpoint = APIOperation
+
+
 class Status(IntEnum):
     """Status of an action or multiple actions."""
 
@@ -683,7 +745,8 @@ class Interaction:
 class TestResult:
     """Result of a single test."""
 
-    endpoint: Endpoint = attr.ib()  # pragma: no mutate
+    method: str = attr.ib()  # pragma: no mutate
+    path: str = attr.ib()  # pragma: no mutate
     data_generation_method: DataGenerationMethod = attr.ib()  # pragma: no mutate
     checks: List[Check] = attr.ib(factory=list)  # pragma: no mutate
     errors: List[Tuple[Exception, Optional[Case]]] = attr.ib(factory=list)  # pragma: no mutate
@@ -738,6 +801,7 @@ class TestResultSet:
     """Set of multiple test results."""
 
     results: List[TestResult] = attr.ib(factory=list)  # pragma: no mutate
+    generic_errors: List[InvalidSchema] = attr.ib(factory=list)  # pragma: no mutate
 
     def __iter__(self) -> Iterator[TestResult]:
         return iter(self.results)
@@ -745,7 +809,7 @@ class TestResultSet:
     @property
     def is_empty(self) -> bool:
         """If the result set contains no results."""
-        return len(self.results) == 0
+        return len(self.results) == 0 and len(self.generic_errors) == 0
 
     @property
     def has_failures(self) -> bool:
@@ -755,7 +819,7 @@ class TestResultSet:
     @property
     def has_errors(self) -> bool:
         """If any result has any errors."""
-        return any(result.has_errors for result in self)
+        return self.errored_count > 0
 
     @property
     def has_logs(self) -> bool:
@@ -775,7 +839,7 @@ class TestResultSet:
 
     @property
     def errored_count(self) -> int:
-        return self._count(lambda result: result.has_errors or result.is_errored)
+        return self._count(lambda result: result.has_errors or result.is_errored) + len(self.generic_errors)
 
     @property
     def total(self) -> Dict[str, Dict[Union[str, Status], int]]:
