@@ -37,6 +37,7 @@ def prepare(
     workers_num: int = 1,
     seed: Optional[int] = None,
     exit_first: bool = False,
+    dry_run: bool = False,
     store_interactions: bool = False,
     fixups: Iterable[str] = (),
     stateful: Optional[Stateful] = None,
@@ -55,7 +56,7 @@ def prepare(
     operation_id: Optional[Filter] = None,
     app: Optional[str] = None,
     validate_schema: bool = True,
-    skip_deprecated_endpoints: bool = False,
+    skip_deprecated_operations: bool = False,
     force_schema_version: Optional[str] = None,
     # Hypothesis-specific configuration
     hypothesis_deadline: Optional[Union[int, NotSet]] = None,
@@ -93,7 +94,7 @@ def prepare(
         operation_id=operation_id,
         app=app,
         validate_schema=validate_schema,
-        skip_deprecated_endpoints=skip_deprecated_endpoints,
+        skip_deprecated_operations=skip_deprecated_operations,
         force_schema_version=force_schema_version,
         checks=checks,
         data_generation_methods=data_generation_methods,
@@ -103,6 +104,7 @@ def prepare(
         seed=seed,
         workers_num=workers_num,
         exit_first=exit_first,
+        dry_run=dry_run,
         auth=auth,
         auth_type=auth_type,
         headers=headers,
@@ -145,7 +147,7 @@ def execute_from_schema(
     operation_id: Optional[Filter] = None,
     app: Optional[str] = None,
     validate_schema: bool = True,
-    skip_deprecated_endpoints: bool = False,
+    skip_deprecated_operations: bool = False,
     force_schema_version: Optional[str] = None,
     checks: Iterable[CheckFunction],
     data_generation_methods: Tuple[DataGenerationMethod, ...] = DEFAULT_DATA_GENERATION_METHODS,
@@ -160,6 +162,7 @@ def execute_from_schema(
     request_tls_verify: Union[bool, str] = True,
     seed: Optional[int] = None,
     exit_first: bool = False,
+    dry_run: bool = False,
     store_interactions: bool = False,
     fixups: Iterable[str] = (),
     stateful: Optional[Stateful] = None,
@@ -185,7 +188,7 @@ def execute_from_schema(
             loader=loader,
             app=app,
             validate_schema=validate_schema,
-            skip_deprecated_endpoints=skip_deprecated_endpoints,
+            skip_deprecated_operations=skip_deprecated_operations,
             auth=auth,
             auth_type=auth_type,
             headers=headers,
@@ -215,6 +218,7 @@ def execute_from_schema(
                     request_timeout=request_timeout,
                     request_tls_verify=request_tls_verify,
                     exit_first=exit_first,
+                    dry_run=dry_run,
                     store_interactions=store_interactions,
                     stateful=stateful,
                     stateful_recursion_limit=stateful_recursion_limit,
@@ -231,6 +235,7 @@ def execute_from_schema(
                     headers=headers,
                     seed=seed,
                     exit_first=exit_first,
+                    dry_run=dry_run,
                     store_interactions=store_interactions,
                     stateful=stateful,
                     stateful_recursion_limit=stateful_recursion_limit,
@@ -249,6 +254,7 @@ def execute_from_schema(
                     seed=seed,
                     workers_num=workers_num,
                     exit_first=exit_first,
+                    dry_run=dry_run,
                     store_interactions=store_interactions,
                     stateful=stateful,
                     stateful_recursion_limit=stateful_recursion_limit,
@@ -269,6 +275,7 @@ def execute_from_schema(
                     request_timeout=request_timeout,
                     request_tls_verify=request_tls_verify,
                     exit_first=exit_first,
+                    dry_run=dry_run,
                     store_interactions=store_interactions,
                     stateful=stateful,
                     stateful_recursion_limit=stateful_recursion_limit,
@@ -285,6 +292,7 @@ def execute_from_schema(
                     headers=headers,
                     seed=seed,
                     exit_first=exit_first,
+                    dry_run=dry_run,
                     store_interactions=store_interactions,
                     stateful=stateful,
                     stateful_recursion_limit=stateful_recursion_limit,
@@ -301,6 +309,7 @@ def execute_from_schema(
                     headers=headers,
                     seed=seed,
                     exit_first=exit_first,
+                    dry_run=dry_run,
                     store_interactions=store_interactions,
                     stateful=stateful,
                     stateful_recursion_limit=stateful_recursion_limit,
@@ -318,7 +327,7 @@ def load_schema(
     loader: Callable = loaders.from_uri,
     app: Any = None,
     validate_schema: bool = True,
-    skip_deprecated_endpoints: bool = False,
+    skip_deprecated_operations: bool = False,
     data_generation_methods: Tuple[DataGenerationMethod, ...] = DEFAULT_DATA_GENERATION_METHODS,
     force_schema_version: Optional[str] = None,
     request_tls_verify: Union[bool, str] = True,
@@ -347,7 +356,7 @@ def load_schema(
         if file_exists(schema_uri):
             loader = loaders.from_path
         elif app is not None and not urlparse(schema_uri).netloc:
-            # If `schema` is not an existing filesystem path, or a URL then it is considered as an endpoint with
+            # If `schema` is not an existing filesystem path, or a URL then it is considered as a path within
             # the given app
             loader = loaders.get_loader_for_app(app)
             loader_options.update(dict_true_values(headers=headers))
@@ -362,7 +371,7 @@ def load_schema(
     return loader(
         schema_uri,
         validate_schema=validate_schema,
-        skip_deprecated_endpoints=skip_deprecated_endpoints,
+        skip_deprecated_operations=skip_deprecated_operations,
         force_schema_version=force_schema_version,
         **loader_options,
     )
