@@ -46,15 +46,17 @@ class SerializedError:
     exception: str = attr.ib()  # pragma: no mutate
     exception_with_traceback: str = attr.ib()  # pragma: no mutate
     example: Optional[SerializedCase] = attr.ib()  # pragma: no mutate
+    title: Optional[str] = attr.ib()  # pragma: no mutate
 
     @classmethod
     def from_error(
-        cls, exception: Exception, case: Optional[Case], headers: Optional[Dict[str, Any]]
+        cls, exception: Exception, case: Optional[Case], headers: Optional[Dict[str, Any]], title: Optional[str] = None
     ) -> "SerializedError":
         return cls(
             exception=format_exception(exception),
             exception_with_traceback=format_exception(exception, True),
             example=SerializedCase.from_case(case, headers) if case else None,
+            title=title,
         )
 
 
@@ -96,8 +98,8 @@ class SerializedTestResult:
     def from_test_result(cls, result: TestResult) -> "SerializedTestResult":
         formatter = logging.Formatter("[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
         return SerializedTestResult(
-            method=result.endpoint.method.upper(),
-            path=result.endpoint.full_path,
+            method=result.method,
+            path=result.path,
             has_failures=result.has_failures,
             has_errors=result.has_errors,
             has_logs=result.has_logs,
