@@ -3,9 +3,25 @@ import pathlib
 import re
 import sys
 import traceback
+import warnings
 from contextlib import contextmanager
 from json import JSONDecodeError
-from typing import Any, Callable, Dict, Generator, List, NoReturn, Optional, Set, Tuple, Type, Union, overload
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Generator,
+    Generic,
+    List,
+    NoReturn,
+    Optional,
+    Set,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    overload,
+)
 
 import requests
 import yaml
@@ -14,8 +30,8 @@ from requests.auth import HTTPDigestAuth
 from requests.exceptions import InvalidHeader  # type: ignore
 from requests.utils import check_header_validity  # type: ignore
 from werkzeug.wrappers import Response as BaseResponse
+from werkzeug.wrappers.json import JSONMixin
 
-from ._compat import JSONMixin
 from .types import Filter, NotSet, RawAuth
 
 try:
@@ -90,6 +106,7 @@ IGNORED_PATTERNS = (
     "Failed to reproduce exception. Expected:",
     "Flaky example!",
     "Traceback (most recent call last):",
+    "You can reproduce this example by temporarily",
 )
 
 
@@ -123,9 +140,13 @@ def capture_hypothesis_output() -> Generator[List[str], None, None]:
 
 
 def format_exception(error: Exception, include_traceback: bool = False) -> str:
+    """Format exception as text."""
+    error_type = type(error)
     if include_traceback:
-        return "".join(traceback.format_exception(type(error), error, error.__traceback__))
-    return "".join(traceback.format_exception_only(type(error), error))
+        lines = traceback.format_exception(error_type, error, error.__traceback__)
+    else:
+        lines = traceback.format_exception_only(error_type, error)
+    return "".join(lines)
 
 
 def parse_content_type(content_type: str) -> Tuple[str, str]:
@@ -138,6 +159,20 @@ def parse_content_type(content_type: str) -> Tuple[str, str]:
     return main_type.lower(), sub_type.lower()
 
 
+def is_json_media_type(value: str) -> bool:
+    """Detect whether the content type is JSON-compatible.
+
+    For example - ``application/problem+json`` matches.
+    """
+    main, sub = parse_content_type(value)
+    return main == "application" and (sub == "json" or sub.endswith("+json"))
+
+
+def is_plain_text_media_type(value: str) -> bool:
+    """Detect variations of the ``text/plain`` media type."""
+    return parse_content_type(value) == ("text", "plain")
+
+
 def are_content_types_equal(source: str, target: str) -> bool:
     """Check if two content types are the same excluding options."""
     return parse_content_type(source) == parse_content_type(target)
@@ -237,6 +272,48 @@ def traverse_schema(schema: Schema, callback: Callable[..., Dict[str, Any]], *ar
         for key, sub_item in schema.items():
             schema[key] = traverse_schema(sub_item, callback, *args, **kwargs)
     elif isinstance(schema, list):
-        for idx, sub_item in enumerate(schema):
-            schema[idx] = traverse_schema(sub_item, callback, *args, **kwargs)
+        schema = [traverse_schema(sub_item, callback, *args, **kwargs) for sub_item in schema]
     return schema
+
+
+def deprecated_property(*, removed_in: str, replacement: str) -> Callable:
+    def wrapper(prop: Callable) -> Callable:
+        @property  # type: ignore
+        def inner(self: Any) -> Any:
+            warnings.warn(
+                f"Property `{prop.__name__}` is deprecated and will be removed in Schemathesis {removed_in}. "
+                f"Use `{replacement}` instead.",
+                DeprecationWarning,
+            )
+            return prop(self)
+
+        return inner
+
+    return wrapper
+
+
+T = TypeVar("T")
+E = TypeVar("E", bound=Exception)
+
+
+class Ok(Generic[T]):
+    __slots__ = ("_value",)
+
+    def __init__(self, value: T):
+        self._value = value
+
+    def ok(self) -> T:
+        return self._value
+
+
+class Err(Generic[E]):
+    __slots__ = ("_error",)
+
+    def __init__(self, error: E):
+        self._error = error
+
+    def err(self) -> E:
+        return self._error
+
+
+Result = Union[Ok[T], Err[E]]
